@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2003-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2003-2024. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -19,6 +19,140 @@
 %%
 
 -module(ct_telnet).
+-moduledoc """
+Common Test specific layer on top of Telnet client ct_telnet_client.erl
+
+`Common Test` specific layer on top of Telnet client `ct_telnet_client.erl`.
+
+Use this module to set up Telnet connections, send commands, and perform string
+matching on the result. For information about how to use `ct_telnet` and
+configure connections, specifically for UNIX hosts, see the `m:unix_telnet`
+manual page.
+
+Default values defined in `ct_telnet`:
+
+[](){: #Default_values }
+
+- Connection timeout (time to wait for connection) = 10 seconds
+- Command timeout (time to wait for a command to return) = 10 seconds
+- Max number of reconnection attempts = 3
+- Reconnection interval (time to wait in between reconnection attempts) = 5
+  seconds
+- Keep alive (sends NOP to the server every 8 sec if connection is idle) =
+  `true`
+- Polling limit (max number of times to poll to get a remaining string
+  terminated) = 0
+- Polling interval (sleep time between polls) = 1 second
+- The TCP_NODELAY option for the telnet socket is disabled (set to `false`) per
+  default
+
+These parameters can be modified by the user with the following configuration
+term:
+
+```erlang
+{telnet_settings, [{connect_timeout,Millisec},
+                   {command_timeout,Millisec},
+                   {reconnection_attempts,N},
+                   {reconnection_interval,Millisec},
+                   {keep_alive,Bool},
+                   {poll_limit,N},
+                   {poll_interval,Millisec},
+                   {tcp_nodelay,Bool}]}.
+```
+
+`Millisec = integer(), N = integer()`
+
+Enter the `telnet_settings` term in a configuration file included in the test
+and `ct_telnet` retrieves the information automatically.
+
+`keep_alive` can be specified per connection, if necessary. For details, see
+`m:unix_telnet`.
+
+## Logging
+
+[](){: #Logging }
+
+The default logging behavior of `ct_telnet` is to print information about
+performed operations, commands, and their corresponding results to the test case
+HTML log. The following is not printed to the HTML log: text strings sent from
+the Telnet server that are not explicitly received by a `ct_telnet` function,
+such as [`expect/3`](`expect/3`). However, `ct_telnet` can be configured to use
+a special purpose event handler, implemented in `ct_conn_log_h`, for logging
+_all_ Telnet traffic. To use this handler, install a `Common Test` hook named
+`cth_conn_log`. Example (using the test suite information function):
+
+```erlang
+suite() ->
+    [{ct_hooks, [{cth_conn_log, [{conn_mod(),hook_options()}]}]}].
+```
+
+`conn_mod()` is the name of the `Common Test` module implementing the connection
+protocol, that is, `ct_telnet`.
+
+The `cth_conn_log` hook performs unformatted logging of Telnet data to a
+separate text file. All Telnet communication is captured and printed, including
+any data sent from the server. The link to this text file is located at the top
+of the test case HTML log.
+
+By default, data for all Telnet connections is logged in one common file (named
+`default`), which can get messy, for example, if multiple Telnet sessions are
+running in parallel. Therefore a separate log file can be created for each
+connection. To configure this, use hook option `hosts` and list the names of the
+servers/connections to be used in the suite. The connections must be named for
+this to work (see [`ct_telnet:open/1,2,3,4`](`open/1`)).
+
+Hook option `log_type` can be used to change the `cth_conn_log` behavior. The
+default value of this option is `raw`, which results in the behavior described
+above. If the value is set to `html`, all Telnet communication is printed to the
+test case HTML log instead.
+
+All `cth_conn_log` hook options described can also be specified in a
+configuration file with configuration variable `ct_conn_log`.
+
+_Example:_
+
+```erlang
+{ct_conn_log, [{ct_telnet,[{log_type,raw},
+                           {hosts,[key_or_name()]}]}]}
+```
+
+> #### Note {: .info }
+>
+> Hook options specified in a configuration file overwrite any hard-coded hook
+> options in the test suite.
+
+[](){: #Logging_example }
+
+_Logging Example:_
+
+The following `ct_hooks` statement causes printing of Telnet traffic to separate
+logs for the connections `server1` and `server2`. Traffic for any other
+connections is logged in the default Telnet log.
+
+```erlang
+suite() ->
+    [{ct_hooks,
+      [{cth_conn_log, [{ct_telnet,[{hosts,[server1,server2]}]}]}]}].
+```
+
+As previously explained, this specification can also be provided by an entry
+like the following in a configuration file:
+
+```erlang
+{ct_conn_log, [{ct_telnet,[{hosts,[server1,server2]}]}]}.
+```
+
+In this case the `ct_hooks` statement in the test suite can look as follows:
+
+```erlang
+suite() ->
+    [{ct_hooks, [{cth_conn_log, []}]}].
+```
+
+## See Also
+
+`m:unix_telnet`
+""".
 
 -export([open/1, open/2, open/3, open/4, close/1]).
 -export([cmd/2, cmd/3, cmdf/3, cmdf/4, get_data/1, 
@@ -42,6 +176,20 @@
 
 -include("ct_util.hrl").
 
+-doc "For `target_name()`, see module `m:ct`.".
+-type connection() :: handle() | {ct:target_name(), connection_type()} | ct:target_name().
+-type connection_type() :: telnet | ts1 | ts2.
+-doc "Handle for a specific Telnet connection, see module `m:ct`.".
+-type handle() :: ct:handle().
+-doc """
+Regular expression matching all possible prompts for a specific target type.
+`regexp` must not have any groups, that is, when matching, `re:run/3` (in
+STDLIB) must return a list with one single element.
+""".
+-type prompt_regexp() :: string().
+-type send_option() :: {'newline', boolean() | string()}.
+-export_type([connection/0, connection_type/0, handle/0, prompt_regexp/0]).
+
 -record(state,{host,
 	       port,
 	       teln_pid,
@@ -61,9 +209,22 @@
 	       reconn_int=?RECONN_TIMEOUT,
 	       tcp_nodelay=false}).
 
+-doc(#{equiv => open(Name, telnet)}).
+-spec open(Name) -> {'ok', Handle} | {'error', Reason}
+              when Name :: atom(),
+                   Handle :: handle(),
+                   Reason :: term().
 open(Name) ->
     open(Name,telnet).
 
+-doc """
+Opens a Telnet connection to the specified target host.
+""".
+-spec open(Name, ConnType) -> {'ok', Handle} | {'error', Reason}
+              when Name :: atom(),
+                   ConnType :: connection_type(),
+                   Handle :: handle(),
+                   Reason :: term().
 open(Name,ConnType) ->
     case ct_util:get_key_from_name(Name) of
 	{ok, unix} -> % unix host
@@ -74,9 +235,47 @@ open(Name,ConnType) ->
 	    Error
     end.
 
+-doc(#{equiv => open(KeyOrName, ConnType, TargetMod, KeyOrName)}).
+-spec open(KeyOrName, ConnType, TargetMod) -> {'ok', Handle} | {'error', Reason}
+              when KeyOrName :: atom(),
+                   ConnType :: connection_type(),
+                   TargetMod :: module(),
+                   Handle :: handle(),
+                   Reason :: term().
 open(KeyOrName,ConnType,TargetMod) ->
     open(KeyOrName,ConnType,TargetMod,KeyOrName).
 
+-doc """
+Opens a Telnet connection to the specified target host.
+
+The target data must exist in a configuration file. The connection can be
+associated with `Name` and/or the returned `Handle`. To allocate a name for the
+target, use one of the following alternatives:
+
+- `ct:require/2` in a test case
+- A `require` statement in the suite information function (`suite/0`)
+- A `require` statement in a test case information function
+
+If you want the connection to be associated with `Handle` only (if you, for
+example, need to open multiple connections to a host), use `Key`, the
+configuration variable name, to specify the target. Notice that a connection
+without an associated target name can only be closed with the `Handle` value.
+
+`TargetMod` is a module that exports the functions
+`connect(Ip, Port, KeepAlive, Extra)` and `get_prompt_regexp()` for the
+specified `TargetType` (for example, `unix_telnet`).
+
+For `target_name()`, see module `m:ct`.
+
+See also `ct:require/2`.
+""".
+-spec open(KeyOrName, ConnType, TargetMod, Extra) -> {'ok', Handle} | {'error', Reason}
+              when KeyOrName :: atom(),
+                   ConnType :: connection_type(),
+                   TargetMod :: module(),
+                   Extra :: term(),
+                   Handle :: handle(),
+                   Reason :: term().
 open(KeyOrName,ConnType,TargetMod,Extra) ->
     case ct:get_config({KeyOrName,ConnType}) of
 	undefined ->
@@ -116,6 +315,16 @@ open(KeyOrName,ConnType,TargetMod,Extra) ->
 					{old,true}])
     end.
 
+-doc """
+Closes the Telnet connection and stops the process managing it.
+
+A connection can be associated with a target name and/or a handle. If
+`Connection` has no associated target name, it can only be closed with the
+handle value (see [`ct_telnet:open/4`](`open/4`)).
+""".
+-spec close(Connection) -> 'ok' | {'error', Reason}
+              when Connection :: connection(),
+                   Reason :: 'already_closed' | term().
 close(Connection) ->
     case get_handle(Connection) of
 	{ok,Pid} ->
@@ -134,9 +343,38 @@ close(Connection) ->
 %%% Test suite interface
 %%%-----------------------------------------------------------------
 
+-doc(#{equiv => cmd(Connection, Cmd, [])}).
+-spec cmd(Connection, Cmd) -> {'ok', Data} | {'error', Reason}
+              when Connection :: connection(),
+                   Cmd :: iodata(),
+                   Data :: string(),
+                   Reason :: term().
 cmd(Connection,Cmd) ->
     cmd(Connection,Cmd,[]).
 
+-doc """
+Sends a command through Telnet and waits for prompt.
+
+By default, this function adds "\\n" to the end of the specified command. If
+this is not desired, use option `{newline,false}`. This is necessary, for
+example, when sending Telnet command sequences prefixed with character Interpret
+As Command (IAC). Option `{newline,string()}` can also be used if a different
+line end than "\\n" is required, for instance `{newline,"\r\n"}`, to add both
+carriage return and newline characters.
+
+Option `timeout` specifies how long the client must wait for prompt. If the time
+expires, the function returns `{error,timeout}`. For information about the
+default value for the command timeout, see the
+[list of default values](`m:ct_telnet#Default_values`) in the beginning of this
+module.
+""".
+-spec cmd(Connection, Cmd, Opts) -> {'ok', Data} | {'error', Reason}
+              when Connection :: connection(),
+                   Cmd :: iodata(),
+                   Opts :: [{'timeout', Timeout} | send_option()] | Timeout,
+                   Timeout :: integer(),
+                   Data :: string(),
+                   Reason :: term().
 cmd(Connection,Cmd,Opts) when is_list(Opts) ->
     case check_cmd_opts(Opts) of
 	ok ->
@@ -161,13 +399,49 @@ check_cmd_opts([]) ->
 check_cmd_opts(Opts) ->
     check_send_opts(Opts).
 
+-doc(#{equiv => cmdf(Connection, CmdFormat, Args, [])}).
+-spec cmdf(Connection, CmdFormat, Args) -> {'ok', Data} | {'error', Reason}
+              when Connection :: connection(),
+                   CmdFormat :: io:format(),
+                   Args :: [term()],
+                   Data :: string(),
+                   Reason :: term().
 cmdf(Connection,CmdFormat,Args) ->
     cmdf(Connection,CmdFormat,Args,[]).
 
+-doc """
+Sends a Telnet command and waits for prompt (uses a format string and a list of
+arguments to build the command).
+
+For details, see [`ct_telnet:cmd/3`](`cmd/3`).
+""".
+-spec cmdf(Connection, CmdFormat, Args, Opts) -> {'ok', Data} | {'error', Reason}
+              when Connection :: connection(),
+                   CmdFormat :: io:format(),
+                   Args :: [term()],
+                   Opts :: [{'timeout', Timeout} | send_option()] | Timeout,
+                   Timeout :: integer(),
+                   Data :: string(),
+                   Reason :: term().
 cmdf(Connection,CmdFormat,Args,Opts) when is_list(Args) ->
     Cmd = lists:flatten(io_lib:format(CmdFormat,Args)),
     cmd(Connection,Cmd,Opts).
 
+-doc """
+Gets all data received by the Telnet client since the last command was sent.
+Only newline-terminated strings are returned. If the last received string has
+not yet been terminated, the connection can be polled automatically until the
+string is complete.
+
+The polling feature is controlled by the configuration values `poll_limit` and
+`poll_interval` and is by default disabled. This means that the function
+immediately returns all complete strings received and saves a remaining
+non-terminated string for a later `get_data` call.
+""".
+-spec get_data(Connection) -> {'ok', Data} | {'error', Reason}
+              when Connection :: connection(),
+                   Data :: string(),
+                   Reason :: term().
 get_data(Connection) ->
     case get_handle(Connection) of
 	{ok,Pid} ->
@@ -176,9 +450,33 @@ get_data(Connection) ->
 	    Error
     end.
 
+-doc(#{equiv => send(Connection, Cmd, [])}).
+-spec send(Connection, Cmd) -> 'ok' | {'error', Reason}
+              when Connection :: connection(),
+                   Cmd :: iodata(),
+                   Reason :: term().
 send(Connection,Cmd) ->
     send(Connection,Cmd,[]).
 
+-doc """
+Sends a Telnet command and returns immediately.
+
+By default, this function adds "\\n" to the end of the specified command. If
+this is not desired, option `{newline,false}` can be used. This is necessary,
+for example, when sending Telnet command sequences prefixed with character
+Interpret As Command (IAC). Option `{newline,string()}` can also be used if a
+different line end than "\\n" is required, for instance `{newline,"\r\n"}`, to
+add both carriage return and newline characters.
+
+The resulting output from the command can be read with
+[`ct_telnet:get_data/2`](`get_data/1`) or [`ct_telnet:expect/2,3`](`expect/2`).
+""".
+-doc(#{since => <<"OTP 17.4">>}).
+-spec send(Connection, Cmd, Opts) -> 'ok' | {'error', Reason}
+              when Connection :: connection(),
+                   Cmd :: iodata(),
+                   Opts :: [send_option()],
+                   Reason :: term().
 send(Connection,Cmd,Opts) ->
     case check_send_opts(Opts) of
 	ok ->
@@ -208,16 +506,150 @@ check_send_opts([Invalid|_]) ->
 check_send_opts([]) ->
     ok.
 
+-doc(#{equiv => sendf(Connection, CmdFormat, Args, [])}).
+-spec sendf(Connection, CmdFormat, Args) -> 'ok' | {'error', Reason}
+              when Connection :: connection(),
+                   CmdFormat :: io:format(),
+                   Args :: [term()],
+                   Reason :: term().
 sendf(Connection,CmdFormat,Args) when is_list(Args) ->
     sendf(Connection,CmdFormat,Args,[]).
 
+-doc """
+Sends a Telnet command and returns immediately (uses a format string and a list
+of arguments to build the command).
+
+For details, see [`ct_telnet:send/3`](`send/3`).
+""".
+-doc(#{since => <<"OTP 17.4">>}).
+-spec sendf(Connection, CmdFormat, Args, Opts) -> 'ok' | {'error', Reason}
+              when Connection :: connection(),
+                   CmdFormat :: io:format(),
+                   Args :: [term()],
+                   Opts :: [send_option()],
+                   Reason :: term().
 sendf(Connection,CmdFormat,Args,Opts) when is_list(Args) ->
     Cmd = lists:flatten(io_lib:format(CmdFormat,Args)),
     send(Connection,Cmd,Opts).
 
+-doc(#{equiv => expect(Connection, Patterns, [])}).
+-spec expect(Connection, Patterns) -> {'ok', Match} | {'ok', Match, HaltReason}
+              | {'error', Reason} | no_return()
+              when Connection :: connection(),
+                   Patterns :: Pattern | [Pattern],
+                   Pattern :: PatternString | {Tag, PatternString},
+                   Tag :: term(),
+                   PatternString :: unicode:charlist(),
+                   Match :: MatchResult | [MatchResult],
+                   MatchResult :: [CaptureData | {Tag, CaptureData}]
+                                | [[CaptureData | {Tag, CaptureData}]],
+                   CaptureData :: string()
+                                | {'error', string(), binary()}
+                                | {'incomplete', string(), binary()},
+                   HaltReason :: MatchResult,
+                   Reason :: term().
 expect(Connection,Patterns) ->
     expect(Connection,Patterns,[]).
 
+-doc """
+Gets data from Telnet and waits for the expected pattern.
+
+`Pattern` can be a POSIX regular expression. The function returns when a pattern
+is successfully matched (at least one, in the case of multiple patterns).
+
+`RxMatch` is a list of matched strings. It looks as follows
+`[FullMatch, SubMatch1, SubMatch2, ...]`, where `FullMatch` is the string
+matched by the whole regular expression, and `SubMatchN` is the string that
+matched subexpression number `N`. Subexpressions are denoted with `'(' ')'` in
+the regular expression.
+
+If a `Tag` is specified, the returned `Match` also includes the matched `Tag`.
+Otherwise, only `RxMatch` is returned.
+
+_Options:_
+
+- **`idle_timeout`** - Indicates that the function must return if the Telnet
+  client is idle (that is, if no data is received) for more than `IdleTimeout`
+  milliseconds. Default time-out is 10 seconds.
+
+- **`total_timeout`** - Sets a time limit for the complete `expect` operation.
+  After `TotalTimeout` milliseconds, `{error,timeout}` is returned. Default is
+  `infinity` (that is, no time limit).
+
+- **`ignore_prompt | no_prompt_check`** - >The function returns when a prompt is
+  received, even if no pattern has yet been matched, and
+  `{error,{prompt,Prompt}}` is returned. However, this behavior can be modified
+  with option `ignore_prompt` or option `no_prompt_check`, which tells `expect`
+  to return only when a match is found or after a time-out.
+
+- **`ignore_prompt`** - `ct_telnet` ignores any prompt found. This option is
+  useful if data sent by the server can include a pattern matching prompt
+  `regexp` (as returned by `TargedMod:get_prompt_regexp/0`), but is not to not
+  cause the function to return.
+
+- **`no_prompt_check`** - `ct_telnet` does not search for a prompt at all. This
+  is useful if, for example, `Pattern` itself matches the prompt.
+
+- **`wait_for_prompt`** - Forces `ct_telnet` to wait until the prompt string is
+  received before returning (even if a pattern has already been matched). This
+  is equal to calling
+  [`expect(Conn, Patterns++[{prompt,Prompt}], [sequence|Opts])`](`expect/3`).
+  Notice that option `idle_timeout` and `total_timeout` can abort the operation
+  of waiting for prompt.
+
+- **`repeat | repeat, N`** - The pattern(s) must be matched multiple times. If
+  `N` is specified, the pattern(s) are matched `N` times, and the function
+  returns `HaltReason = done`. This option can be interrupted by one or more
+  `HaltPatterns`. `MatchList` is always returned, that is, a list of `Match`
+  instead of only one `Match`. Also `HaltReason` is returned.
+
+- **`sequence`** - All patterns must be matched in a sequence. A match is not
+  concluded until all patterns are matched. This option can be interrupted by
+  one or more `HaltPatterns`. `MatchList` is always returned, that is, a list of
+  `Match` instead of only one `Match`. Also `HaltReason` is returned.
+
+_Example 1:_
+
+```erlang
+expect(Connection,[{abc,"ABC"},{xyz,"XYZ"}],[sequence,{halt,[{nnn,"NNN"}]}])
+```
+
+First this tries to match `"ABC"`, and then `"XYZ"`, but if `"NNN"` appears, the
+function returns `{error,{nnn,["NNN"]}}`. If both `"ABC"` and `"XYZ"` are
+matched, the function returns `{ok,[AbcMatch,XyzMatch]}`.
+
+_Example 2:_
+
+```erlang
+expect(Connection,[{abc,"ABC"},{xyz,"XYZ"}],[{repeat,2},{halt,[{nnn,"NNN"}]}])
+```
+
+This tries to match `"ABC"` or `"XYZ"` twice. If `"NNN"` appears, the function
+returns `HaltReason = {nnn,["NNN"]}`.
+
+Options `repeat` and `sequence` can be combined to match a sequence multiple
+times.
+""".
+-spec expect(Connection, Patterns, Opts) -> {'ok', Match} | {'ok', Match, HaltReason}
+              | {'error', Reason} | no_return()
+              when Connection :: connection(),
+                   Patterns :: Pattern | [Pattern],
+                   Pattern :: PatternString | {Tag, PatternString},
+                   Tag :: term(),
+                   PatternString :: unicode:charlist(),
+                   Opts :: [Option],
+                   Option :: {'idle_timeout', non_neg_integer()}
+                           | {'total_timeout', non_neg_integer()} | 'ignore_prompt'
+                           | 'no_prompt_check' | 'wait_for_prompt' | 'repeat'
+                           | {'repeat', non_neg_integer()} | 'sequence' | {'halt', Patterns},
+                   Match :: MatchResult | [MatchResult],
+                   MatchResult :: [CaptureData | {Tag, CaptureData}]
+                                | [[CaptureData | {Tag, CaptureData}]],
+                   CaptureData :: string()
+                                | {'error', string(), binary()}
+                                | {'incomplete', string(), binary()},
+                   HaltReason :: MatchResult,
+                   Reason :: term().
 expect(Connection,Patterns,Opts) ->
     case get_handle(Connection) of
         {ok,Pid} ->
@@ -235,6 +667,7 @@ expect(Connection,Patterns,Opts) ->
 %%%=================================================================
 %%% Callback functions
 
+-doc false.
 init(Name,{Ip,Port,Type},{TargetMod,KeepAlive,Extra}) ->
     S0 = case ct:get_config(telnet_settings) of
 	     undefined ->
@@ -315,6 +748,7 @@ set_telnet_defaults([Unknown|Ss],S) ->
 set_telnet_defaults([],S) ->
     S.
 
+-doc false.
 handle_msg({cmd,Cmd,Opts},State) ->
     start_gen_log(heading(cmd,State#state.name)),
     log(State,cmd,"Cmd: ~tp",[Cmd]),
@@ -440,6 +874,7 @@ handle_msg({expect,Pattern,Opts},State) ->
     {Return1,State#state{buffer=NewBuffer,prompt=Prompt}}.
 
 
+-doc false.
 reconnect({Ip,Port,_Type},State) ->
     reconnect(Ip,Port,State#state.reconns,State).
 reconnect(Ip,Port,N,State=#state{name=Name,
@@ -470,6 +905,7 @@ reconnect(Ip,Port,N,State=#state{name=Name,
     end.
 
 
+-doc false.
 terminate(TelnPid,State) ->
     Result = ct_telnet_client:close(TelnPid),
     log(State,close,"Telnet connection for ~w closed.",[TelnPid]),
@@ -537,6 +973,7 @@ force_log(State,Action,String,Args) ->
     log(State,Action,String,Args,true).
 
 %%%-----------------------------------------------------------------
+-doc false.
 log(State,Action,String,Args) when is_record(State, state) ->
     log(State,Action,String,Args,false);
 log(Name,Action,String,Args) when is_atom(Name) ->
@@ -545,6 +982,7 @@ log(TelnPid,Action,String,Args) when is_pid(TelnPid) ->
     log(#state{teln_pid=TelnPid},Action,String,Args,false).
 
 %%%-----------------------------------------------------------------
+-doc false.
 log(undefined,String,Args) ->
     log(#state{},undefined,String,Args,false);
 log(Name,String,Args) when is_atom(Name) ->
@@ -603,6 +1041,7 @@ log(#state{name=Name,teln_pid=TelnPid,host=Host,port=Port},
     end.
 
 %%%-----------------------------------------------------------------
+-doc false.
 start_gen_log(Heading) ->
     %% check if output is suppressed
     case ct_util:is_silenced(telnet) of
@@ -611,6 +1050,7 @@ start_gen_log(Heading) ->
     end.
 
 %%%-----------------------------------------------------------------
+-doc false.
 end_gen_log() -> 
     %% check if output is suppressed
     case ct_util:is_silenced(telnet) of
@@ -625,6 +1065,7 @@ debug_cont_gen_log(Str,Args) ->
     put(silent,Old).
 
 %% Log callback - called from the error handler process
+-doc false.
 format_data(_How,{String,Args}) ->
     io_lib:format(String,Args).
 
@@ -670,6 +1111,7 @@ teln_get_all_data(State=#state{teln_pid=Pid,prx=Prx},Data,Acc,LastLine,Polls) ->
 %% Externally the silent_teln_expect function shall only be used
 %% by the TargetModule, i.e. the target specific module which
 %% implements connect/2 and get_prompt_regexp/0.
+-doc false.
 silent_teln_expect(Name,Pid,Data,Pattern,Prx,Opts) ->
     Old = put(silent,true),
     Result = teln_expect(Name,Pid,Data,Pattern,Prx,Opts),
@@ -1019,7 +1461,7 @@ seq_expect(Name,Pid,Data,Patterns,Acc,EO) ->
     end.
 
 %% seq_expect1: For one prompt-chunk, match each pattern - line by
-%% line if it is other than the prompt we are seaching for.
+%% line if it is other than the prompt we are searching for.
 seq_expect1(Name,Pid,Data,[prompt|Patterns],Acc,Rest,EO) ->
     case EO#eo.found_prompt of
 	false ->
@@ -1185,6 +1627,7 @@ add_tabs([],[$\n|Acc],LastLine) ->
 add_tabs([],[],LastLine) ->
     {[],lists:reverse(LastLine)}.
 
+-doc false.
 teln_receive_until_prompt(Pid,Prx,Timeout) ->
     Fun = fun() -> teln_receive_until_prompt(Pid,Prx,[],[]) end,
     ct_gen_conn:do_within_time(Fun, Timeout).
@@ -1224,37 +1667,36 @@ split_lines([Char|Rest],Line,Lines) ->
 split_lines([],Line,Lines) ->
     {Lines,lists:reverse(Line)}.
 
-
-match_prompt(Str,Prx) ->
-    match_prompt(Str,Prx,[]).
-match_prompt(Str,Prx,Acc) ->
+match_prompt(Str, Prx) ->
+    match_prompt(unicode:characters_to_binary(Str), Prx, []).
+match_prompt(Str, Prx, Acc) ->
     case re:run(Str,Prx,[unicode]) of
-	nomatch ->
-	    noprompt;
-	{match,[{Start,Len}]} ->
-	    case split_prompt_string(Str,Start+1,Start+Len,1,[],[]) of
-		{noprompt,Done,Rest} ->
-		    match_prompt(Rest,Prx,Done);
-		{prompt,UptoPrompt,Prompt,Rest} ->
-		    {prompt,lists:reverse(UptoPrompt++Acc),
-		     lists:reverse(Prompt),Rest}
-	    end
+        nomatch ->
+            noprompt;
+        {match,[{Start,Len}]} ->
+            <<UptoPrompt:Start/binary, Prompt:Len/binary, Rest/binary>> = Str,
+            case validate_prompt(Start, UptoPrompt, Prompt) of
+                ok ->
+                    {prompt,
+                     unicode:characters_to_list([lists:reverse(Acc), UptoPrompt, Prompt]),
+                     unicode:characters_to_list(Prompt),
+                     unicode:characters_to_list(Rest)};
+                recurse ->
+                    <<Skip:(Start+Len)/binary, Cont/binary>> = Str,
+                    match_prompt(Cont, Prx, [Skip|Acc])
+            end
     end.
 
-split_prompt_string([Ch|Str],Start,End,N,UptoPrompt,Prompt) when N<Start ->
-    split_prompt_string(Str,Start,End,N+1,[Ch|UptoPrompt],Prompt);
-split_prompt_string([Ch|Str],Start,End,N,UptoPrompt,Prompt) 
-  when N>=Start, N<End->
-    split_prompt_string(Str,Start,End,N+1,UptoPrompt,[Ch|Prompt]);
-split_prompt_string([Ch|Rest],_Start,End,N,UptoPrompt,Prompt) when N==End ->
-    case UptoPrompt of
-	[$",$=,$T,$P,$M,$O,$R,$P|_] ->
-	    %% This is a line from "listenv", it is not a real prompt
-	    {noprompt,[Ch|Prompt]++UptoPrompt,Rest};
-	[$\s,$t,$s,$a|_] when Prompt==":nigol" ->
-	    %% This is probably the "Last login:" statement which is
-	    %% written when telnet connection is openend.
-	    {noprompt,[Ch|Prompt]++UptoPrompt,Rest};
-	_ ->
-	    {prompt,[Ch|Prompt]++UptoPrompt,[Ch|Prompt],Rest}
+validate_prompt(Size, PrePrompt,  Prompt) ->
+    case PrePrompt of
+        %% This is a line from "listenv", it is not a real prompt
+        <<_:(Size-8)/binary, "PROMPT=\"", _/binary>> ->
+            recurse;
+        %% This is probably the "Last login:" statement which is
+        %% written when telnet connection is opened.
+        <<_:(Size-5)/binary, _L:8, "ast ", _/binary>>
+          when Prompt =:= <<"login: ">> ->
+            recurse;
+        _ ->
+            ok
     end.
