@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2000-2018. All Rights Reserved.
+%% Copyright Ericsson AB 2000-2024. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -18,8 +18,137 @@
 %% %CopyrightEnd%
 %%
 -module(beam_lib).
+-moduledoc """
+This module provides an interface to files created by the BEAM Compiler ("BEAM
+files").
+
+The format used, a variant of "EA IFF 1985" Standard for Interchange Format Files,
+divides data into chunks.
+
+Chunk data can be returned as binaries or as compound terms. Compound terms are
+returned when chunks are referenced by names (atoms) rather than identifiers
+(strings). The recognized names and the corresponding identifiers are as
+follows:
+
+- `atoms ("Atom")`
+- `attributes ("Attr")`
+- `compile_info ("CInf")`
+- `debug_info ("Dbgi")`
+- `exports ("ExpT")`
+- `imports ("ImpT")`
+- `indexed_imports ("ImpT")`
+- `labeled_exports ("ExpT")`
+- `labeled_locals ("LocT")`
+- `locals ("LocT")`
+- `documentation ("Docs")`
+
+[](){: #debug_info }
+
+## Debug Information/Abstract Code
+
+Option `debug_info` can be specified to the Compiler (see
+[`compile`](`m:compile#debug_info`)) to have debug information, such as
+[Erlang Abstract Format](`e:erts:absform.md`), stored in the `debug_info` chunk.
+Tools such as Debugger and Xref require the debug information to be included.
+
+> #### Warning {: .warning }
+>
+> Source code can be reconstructed from the debug information. To prevent this,
+> use encrypted debug information (see below).
+
+The debug information can also be removed from BEAM files using `strip/1`,
+`strip_files/1`, and/or `strip_release/1`.
+
+## Reconstruct Source Code
+
+The following example shows how to reconstruct Erlang source code from the debug
+information in a BEAM file `Beam`:
+
+```erlang
+{ok,{_,[{abstract_code,{_,AC}}]}} = beam_lib:chunks(Beam,[abstract_code]).
+io:fwrite("~s~n", [erl_prettypr:format(erl_syntax:form_list(AC))]).
+```
+
+## Encrypted Debug Information
+
+The debug information can be encrypted to keep the source code secret, but still
+be able to use tools such as Debugger or Xref.
+
+To use encrypted debug information, a key must be provided to the compiler and
+`beam_lib`. The key is specified as a string. It is recommended that the string
+contains at least 32 characters and that both upper and lower case letters as
+well as digits and special characters are used.
+
+The default type (and currently the only type) of crypto algorithm is
+`des3_cbc`, three rounds of DES. The key string is scrambled using
+`erlang:md5/1` to generate the keys used for `des3_cbc`.
+
+> #### Note {: .info }
+>
+> As far as we know by the time of writing, it is infeasible to break `des3_cbc`
+> encryption without any knowledge of the key. Therefore, as long as the key is
+> kept safe and is unguessable, the encrypted debug information _should_ be safe
+> from intruders.
+
+The key can be provided in the following two ways:
+
+1. Use Compiler option `{debug_info_key,Key}`, see
+   [`compile`](`m:compile#debug_info_key`) and function `crypto_key_fun/1` to
+   register a fun that returns the key whenever `beam_lib` must decrypt the
+   debug information.
+
+If no such fun is registered, `beam_lib` instead searches for an `.erlang.crypt`
+file, see the next section.
+
+1. Store the key in a text file named `.erlang.crypt`.
+
+In this case, Compiler option `encrypt_debug_info` can be used, see
+[`compile`](`m:compile#encrypt_debug_info`).
+
+## .erlang.crypt
+
+`beam_lib` searches for `.erlang.crypt` in the current directory, then the
+[user's home directory](`m:init#home`) and then
+[`filename:basedir(user_config, "erlang")`](`m:filename#user_config`). If the
+file is found and contains a key, `beam_lib` implicitly creates a crypto key fun
+and registers it.
+
+File `.erlang.crypt` is to contain a single list of tuples:
+
+```erlang
+{debug_info, Mode, Module, Key}
+```
+
+`Mode` is the type of crypto algorithm; currently, the only allowed value is
+`des3_cbc`. `Module` is either an atom, in which case `Key` is only used for the
+module `Module`, or `[]`, in which case `Key` is used for all modules. `Key` is
+the non-empty key string.
+
+`Key` in the first tuple where both `Mode` and `Module` match is used.
+
+The following is an example of an `.erlang.crypt` file that returns the same key
+for all modules:
+
+```erlang
+[{debug_info, des3_cbc, [], "%>7}|pc/DM6Cga*68$Mw]L#&_Gejr]G^"}].
+```
+
+The following is a slightly more complicated example of an `.erlang.crypt`
+providing one key for module `t` and another key for all other modules:
+
+```erlang
+[{debug_info, des3_cbc, t, "My KEY"},
+ {debug_info, des3_cbc, [], "%>7}|pc/DM6Cga*68$Mw]L#&_Gejr]G^"}].
+```
+
+> #### Note {: .info }
+>
+> Do not use any of the keys in these examples. Use your own keys.
+""".
 -behaviour(gen_server).
 
+-include_lib("kernel/include/eep48.hrl").
+
 %% Avoid warning for local function error/1 clashing with autoimported BIF.
 -compile({no_auto_import,[error/1]}).
 %% Avoid warning for local function error/2 clashing with autoimported BIF.
@@ -32,8 +161,12 @@
 	 all_chunks/1,
 	 diff_dirs/2,
 	 strip/1,
+	 strip/2,
 	 strip_files/1,
+	 strip_files/2,
 	 strip_release/1,
+	 strip_release/2,
+	 significant_chunks/0,
 	 build_module/1,
 	 version/1,
 	 md5/1,
@@ -47,36 +180,77 @@
 -export([make_crypto_key/2, get_crypto_key/1]).	%Utilities used by compiler
 
 -export_type([attrib_entry/0, compinfo_entry/0, labeled_entry/0, label/0]).
+-export_type([chunkid/0]).
+-export_type([chnk_rsn/0]).
+-export_type([beam/0]).
 
 -import(lists, [append/1, delete/2, foreach/2, keysort/2, 
 		member/2, reverse/1, sort/1, splitwith/2]).
 
 %%-------------------------------------------------------------------------
 
--type beam() :: module() | file:filename() | binary().
+-doc """
+Each of the functions described below accept either the filename (as a string)
+or a binary containing the BEAM module.
+""".
+-type beam() :: file:filename() | binary().
+-doc """
+The format stored in the `debug_info` chunk.
+
+To retrieve particular code representation from the backend,
+`Backend:debug_info(Format, Module, Data, Opts)` must be invoked. `Format` is an
+atom, such as `erlang_v1` for the Erlang Abstract Format or `core_v1` for Core
+Erlang. `Module` is the module represented by the beam file and `Data` is the
+value stored in the debug info chunk. `Opts` is any list of values supported by
+the `Backend`. `Backend:debug_info/4` must return `{ok, Code}` or
+`{error, Term}`.
+
+Developers must always invoke the `debug_info/4` function and never rely on the
+`Data` stored in the `debug_info` chunk, as it is opaque and may change at any
+moment. `no_debug_info` means that chunk `"Dbgi"` is present, but empty.
+""".
 -type debug_info() :: {DbgiVersion :: atom(), Backend :: module(), Data :: term()} | 'no_debug_info'.
 
 -type forms()     :: [erl_parse:abstract_form() | erl_parse:form_info()].
 
+-doc """
+It is not checked that the forms conform to the abstract format indicated by
+`AbstVersion`. `no_abstract_code` means that chunk `"Abst"` is present, but
+empty.
+
+For modules compiled with OTP 20 onwards, the `abst_code` chunk is automatically
+computed from the `debug_info` chunk.
+""".
 -type abst_code() :: {AbstVersion :: atom(), forms()} | 'no_abstract_code'.
 -type dataB()     :: binary().
 -type index()     :: non_neg_integer().
 -type label()     :: integer().
 
+-doc """
+`"Attr" | "CInf" | "Dbgi" | "ExpT" | "ImpT" | "LocT" | "AtU8" | "Docs"`
+""".
 -type chunkid()   :: nonempty_string(). % approximation of the strings below
-%% "Abst" | "Dbgi" | "Attr" | "CInf" | "ExpT" | "ImpT" | "LocT" | "Atom" | "AtU8".
+%% "Abst" | "Dbgi" | "Attr" | "CInf" | "ExpT" | "ImpT" | "LocT" | "Atom" | "AtU8" | "Docs"
 -type chunkname() :: 'abstract_code' | 'debug_info'
                    | 'attributes' | 'compile_info'
                    | 'exports' | 'labeled_exports'
                    | 'imports' | 'indexed_imports'
                    | 'locals' | 'labeled_locals'
-                   | 'atoms'.
+                   | 'atoms' | 'documentation'.
 -type chunkref()  :: chunkname() | chunkid().
 
 -type attrib_entry()   :: {Attribute :: atom(), [AttributeValue :: term()]}.
 -type compinfo_entry() :: {InfoKey :: atom(), term()}.
 -type labeled_entry()  :: {Function :: atom(), arity(), label()}.
 
+-doc "[EEP-48 documentation format](`e:kernel:eep48_chapter.md#the-docs-format`)".
+-type docs() :: #docs_v1{}.
+
+-doc """
+The list of attributes is sorted on `Attribute` (in `t:attrib_entry/0`) and each
+attribute name occurs once in the list. The attribute values occur in the same
+order as in the file. The lists of functions are also sorted.
+""".
 -type chunkdata() :: {chunkid(), dataB()}
                    | {'abstract_code', abst_code()}
                    | {'debug_info', debug_info()}
@@ -88,7 +262,8 @@
                    | {'indexed_imports', [{index(), module(), Function :: atom(), arity()}]}
                    | {'locals', [{atom(), arity()}]}
                    | {'labeled_locals', [labeled_entry()]}
-                   | {'atoms', [{integer(), atom()}]}.
+                   | {'atoms', [{integer(), atom()}]}
+                   | {'documentation', docs()}.
 
 %% Error reasons
 -type info_rsn()  :: {'chunk_too_big', file:filename(),
@@ -103,6 +278,7 @@
 -type chnk_rsn()  :: {'unknown_chunk', file:filename(), atom()}
                    | {'key_missing_or_invalid', file:filename(),
 		      'abstract_code' | 'debug_info'}
+                   | {'missing_backend', file:filename(), module()}
                    | info_rsn().
 -type cmp_rsn()   :: {'modules_different', module(), module()}
                    | {'chunks_different', chunkid()}
@@ -115,6 +291,18 @@
 %%  Exported functions
 %%
 
+-doc """
+Returns a list containing some information about a BEAM file as tuples
+`{Item, Info}`:
+
+- **`{file, Filename} | {binary, Binary}`** - The name (string) of the BEAM
+  file, or the binary from which the information was extracted.
+
+- **`{module, Module}`** - The name (atom) of the module.
+
+- **`{chunks, [{ChunkId, Pos, Size}]}`** - For each chunk, the identifier
+  (string) and the position and size of the chunk data, in bytes.
+""".
 -spec info(Beam) -> [InfoPair] | {'error', 'beam_lib', info_rsn()} when
       Beam :: beam(),
       InfoPair :: {'file', Filename :: file:filename()}
@@ -127,6 +315,10 @@
 info(File) ->
     read_info(beam_filename(File)).
 
+-doc """
+Reads chunk data for selected chunks references. The order of the returned list
+of chunk data is determined by the order of the list of chunks references.
+""".
 -spec chunks(Beam, ChunkRefs) ->
                     {'ok', {module(), [chunkdata()]}} |
                     {'error', 'beam_lib', chnk_rsn()} when
@@ -136,6 +328,17 @@ info(File) ->
 chunks(File, Chunks) ->
     read_chunk_data(File, Chunks).
 
+-doc """
+Reads chunk data for selected chunks references. The order of the returned list
+of chunk data is determined by the order of the list of chunks references.
+
+By default, if any requested chunk is missing in `Beam`, an `error` tuple is
+returned. However, if option `allow_missing_chunks` is specified, a result is
+returned even if chunks are missing. In the result list, any missing chunks are
+represented as `{ChunkRef,missing_chunk}`. Notice however that if chunk `"Atom"`
+is missing, that is considered a fatal error and the return value is an `error`
+tuple.
+""".
 -spec chunks(Beam, ChunkRefs, Options) ->
                     {'ok', {module(), [ChunkResult]}} |
                     {'error', 'beam_lib', chnk_rsn()} when
@@ -148,12 +351,22 @@ chunks(File, Chunks, Options) ->
     try read_chunk_data(File, Chunks, Options)
     catch Error -> Error end.
 
+-doc "Reads chunk data for all chunks.".
+-doc(#{since => <<"OTP 18.2">>}).
 -spec all_chunks(beam()) ->
-           {'ok', 'beam_lib', [{chunkid(), dataB()}]} | {'error', 'beam_lib', info_rsn()}.
+           {'ok', module(), [{chunkid(), dataB()}]} | {'error', 'beam_lib', info_rsn()}.
 
 all_chunks(File) ->
     read_all_chunks(File).
 
+-doc """
+Compares the contents of two BEAM files.
+
+If the module names are the same, and all chunks except for chunk `"CInf"`
+(the chunk containing the compilation information that is returned by
+`Module:module_info(compile)`) have the same contents in both files, `ok` is
+returned. Otherwise an error message is returned.
+""".
 -spec cmp(Beam1, Beam2) -> 'ok' | {'error', 'beam_lib', cmp_rsn()} when
       Beam1 :: beam(),
       Beam2 :: beam().
@@ -162,6 +375,15 @@ cmp(File1, File2) ->
     try cmp_files(File1, File2)
     catch Error -> Error end.
 
+-doc """
+Compares the BEAM files in two directories.
+
+Only files with extension `".beam"` are compared. BEAM files that exist only in
+directory `Dir1` (`Dir2`) are returned in `Only1` (`Only2`). BEAM files that
+exist in both directories but are considered different by [`cmp/2`](`cmp/2`) are
+ returned as pairs \{`Filename1`, `Filename2`\}, where `Filename1` (`Filename2`)
+exists in directory `Dir1` (`Dir2`).
+""".
 -spec cmp_dirs(Dir1, Dir2) ->
            {Only1, Only2, Different} | {'error', 'beam_lib', Reason} when
       Dir1 :: atom() | file:filename(),
@@ -174,6 +396,11 @@ cmp(File1, File2) ->
 cmp_dirs(Dir1, Dir2) ->
     catch compare_dirs(Dir1, Dir2).
 
+-doc """
+Compares the BEAM files in two directories as `cmp_dirs/2`, but the names of
+files that exist in only one directory or are different are presented on
+standard output.
+""".
 -spec diff_dirs(Dir1, Dir2) -> 'ok' | {'error', 'beam_lib', Reason} when
       Dir1 :: atom() | file:filename(),
       Dir2 :: atom() | file:filename(),
@@ -182,24 +409,78 @@ cmp_dirs(Dir1, Dir2) ->
 diff_dirs(Dir1, Dir2) ->
     catch diff_directories(Dir1, Dir2).
 
+-doc """
+Removes all chunks from a BEAM file except those used by the loader.
+
+In particular, the debug information (chunk `debug_info` and `abstract_code`) is
+removed.
+""".
 -spec strip(Beam1) ->
         {'ok', {module(), Beam2}} | {'error', 'beam_lib', info_rsn()} when
       Beam1 :: beam(),
       Beam2 :: beam().
 
 strip(FileName) ->
-    try strip_file(FileName)
+    strip(FileName, []).
+
+-doc """
+Removes all chunks from a BEAM file except those used by the loader or mentioned
+in `AdditionalChunks`.
+
+In particular, the debug information (chunk `debug_info` and `abstract_code`) is removed.
+""".
+-doc(#{since => <<"OTP 22.0">>}).
+-spec strip(Beam1, AdditionalChunks) ->
+        {'ok', {module(), Beam2}} | {'error', 'beam_lib', info_rsn()} when
+      Beam1 :: beam(),
+      AdditionalChunks :: [chunkid()],
+      Beam2 :: beam().
+
+strip(FileName, AdditionalChunks) ->
+    try strip_file(FileName, AdditionalChunks)
     catch Error -> Error end.
     
+-doc """
+Removes all chunks except those used by the loader from `Files`.
+
+In particular, the debug information (chunk `debug_info` and `abstract_code`) is
+removed. The returned list contains one element for each specified filename, in
+the same order as in `Files`.
+""".
 -spec strip_files(Files) ->
         {'ok', [{module(), Beam}]} | {'error', 'beam_lib', info_rsn()} when
       Files :: [beam()],
       Beam :: beam().
 
-strip_files(Files) when is_list(Files) ->
-    try strip_fils(Files)
+strip_files(Files) ->
+    strip_files(Files, []).
+
+-doc """
+Removes all chunks except those used by the loader or mentioned in
+`AdditionalChunks` from `Files`.
+
+In particular, the debug information (chunk `debug_info` and `abstract_code`) is
+removed. The returned list contains one element for each specified filename,
+in the same order as in `Files`.
+""".
+-doc(#{since => <<"OTP 22.0">>}).
+-spec strip_files(Files, AdditionalChunks) ->
+        {'ok', [{module(), Beam}]} | {'error', 'beam_lib', info_rsn()} when
+      Files :: [beam()],
+      AdditionalChunks :: [chunkid()],
+      Beam :: beam().
+
+strip_files(Files, AdditionalChunks) when is_list(Files) ->
+    try strip_fils(Files, AdditionalChunks)
     catch Error -> Error end.
 
+-doc """
+Removes all chunks except those used by the loader from the BEAM files of a
+release.
+
+`Dir` is to be the installation root directory. For example, the current OTP
+release can be stripped with the call `beam_lib:strip_release(code:root_dir())`.
+""".
 -spec strip_release(Dir) ->
         {'ok', [{module(), file:filename()}]}
       | {'error', 'beam_lib', Reason} when
@@ -207,8 +488,48 @@ strip_files(Files) when is_list(Files) ->
       Reason :: {'not_a_directory', term()} | info_rsn().
 
 strip_release(Root) ->
-    catch strip_rel(Root).
+    strip_release(Root, []).
 
+-doc """
+Removes all chunks except those used by the loader or mentioned in
+`AdditionalChunks`.
+
+`Dir` is to be the installation root directory. For example, the current OTP
+release can be stripped with the call `beam_lib:strip_release(code:root_dir(),[documentation])`.
+""".
+-doc(#{since => <<"OTP 22.0">>}).
+-spec strip_release(Dir, AdditionalChunks) ->
+        {'ok', [{module(), file:filename()}]}
+      | {'error', 'beam_lib', Reason} when
+      Dir :: atom() | file:filename(),
+      AdditionalChunks :: [chunkid()],
+      Reason :: {'not_a_directory', term()} | info_rsn().
+
+strip_release(Root, AdditionalChunks) ->
+    catch strip_rel(Root, AdditionalChunks).
+
+-doc """
+Returns the module version or versions. A version is defined by module attribute
+`-vsn(Vsn)`.
+
+If this attribute is not specified, the version defaults to the
+checksum of the module. Notice that if version `Vsn` is not a list, it is made
+into one, that is `{ok,{Module,[Vsn]}}` is returned. If there are many `-vsn`
+module attributes, the result is the concatenated list of versions.
+
+_Examples:_
+
+```erlang
+1> beam_lib:version(a). % -vsn(1).
+{ok,{a,[1]}}
+2> beam_lib:version(b). % -vsn([1]).
+{ok,{b,[1]}}
+3> beam_lib:version(c). % -vsn([1]). -vsn(2).
+{ok,{c,[1,2]}}
+4> beam_lib:version(d). % no -vsn attribute
+{ok,{d,[275613208176997377698094100858909383631]}}
+```
+""".
 -spec version(Beam) ->
                      {'ok', {module(), [Version :: term()]}} |
                      {'error', 'beam_lib', chnk_rsn()} when
@@ -223,6 +544,10 @@ version(File) ->
 	    Error
     end.
 
+-doc """
+Calculates an MD5 redundancy check for the code of the module (compilation date
+and other attributes are not included).
+""".
 -spec md5(Beam) ->
         {'ok', {module(), MD5}} | {'error', 'beam_lib', chnk_rsn()} when
       Beam :: beam(),
@@ -237,6 +562,11 @@ md5(File) ->
 	    Error
     end.
 
+-doc """
+For a specified error returned by any function in this module, this function
+returns a descriptive string of the error in English. For file errors, function
+[`file:format_error(Posix)`](`file:format_error/1`) is to be called.
+""".
 -spec format_error(Reason) -> io_lib:chars() when
       Reason :: term().
 
@@ -278,6 +608,9 @@ format_error(badfun) ->
     "not a fun or the fun has the wrong arity";
 format_error(exists) ->
     "a fun has already been installed";
+format_error({missing_backend, File, Backend}) ->
+    io_lib:format("~tp: Cannot retrieve abstract code because the backend ~p is missing",
+		  [File, Backend]);
 format_error(E) ->
     io_lib:format("~tp~n", [E]).
 
@@ -291,6 +624,42 @@ format_error(E) ->
                         | {'debug_info', mode(), module(), file:filename()}.
 -type crypto_fun()     :: fun((crypto_fun_arg()) -> term()).
 
+-doc """
+Registers an unary fun that is called if `beam_lib` must read an `debug_info`
+chunk that has been encrypted. The fun is held in a process that is started by
+the function.
+
+If a fun is already registered when attempting to register a fun,
+`{error, exists}` is returned.
+
+The fun must handle the following arguments:
+
+```erlang
+CryptoKeyFun(init) -> ok | {ok, NewCryptoKeyFun} | {error, Term}
+```
+
+Called when the fun is registered, in the process that holds the fun. Here the
+crypto key fun can do any necessary initializations. If `{ok, NewCryptoKeyFun}`
+is returned, `NewCryptoKeyFun` is registered instead of `CryptoKeyFun`. If
+`{error, Term}` is returned, the registration is aborted and
+[`crypto_key_fun/1`](`crypto_key_fun/1`) also returns `{error, Term}`.
+
+```erlang
+CryptoKeyFun({debug_info, Mode, Module, Filename}) -> Key
+```
+
+Called when the key is needed for module `Module` in the file named `Filename`.
+`Mode` is the type of crypto algorithm; currently, the only possible value is
+`des3_cbc`. The call is to fail (raise an exception) if no key is available.
+
+```text
+CryptoKeyFun(clear) -> term()
+```
+
+Called before the fun is unregistered. Here any cleaning up can be done. The
+return value is not important, but is passed back to the caller of
+`clear_crypto_key_fun/0` as part of its return value.
+""".
 -spec crypto_key_fun(CryptoKeyFun) -> 'ok' | {'error', Reason} when
       CryptoKeyFun :: crypto_fun(),
       Reason :: badfun | exists | term().
@@ -298,12 +667,21 @@ format_error(E) ->
 crypto_key_fun(F) ->
     call_crypto_server({crypto_key_fun, F}).
 
+-doc """
+Unregisters the crypto key fun and terminates the process holding it, started by
+`crypto_key_fun/1`.
+
+Returns either `{ok, undefined}` if no crypto key fun is registered, or
+`{ok, Term}`, where `Term` is the return value from `CryptoKeyFun(clear)`, see
+[`crypto_key_fun/1`](`crypto_key_fun/1`).
+""".
 -spec clear_crypto_key_fun() -> 'undefined' | {'ok', Result} when
       Result :: 'undefined' | term().
 
 clear_crypto_key_fun() ->
     call_crypto_server(clear_crypto_key_fun).
 
+-doc false.
 -spec make_crypto_key(mode(), string()) ->
         {mode(), [binary()], binary(), integer()}.
 
@@ -312,6 +690,8 @@ make_crypto_key(des3_cbc=Type, String) ->
     <<K3:8/binary,IVec:8/binary>> = erlang:md5([First|reverse(String)]),
     {Type,[K1,K2,K3],IVec,8}.
 
+-doc "Builds a BEAM module (as a binary) from a list of chunks.".
+-doc(#{since => <<"OTP 18.2">>}).
 -spec build_module(Chunks) -> {'ok', Binary} when
       Chunks :: [{chunkid(), dataB()}],
       Binary :: binary().
@@ -401,17 +781,17 @@ cmp_lists([{Id, C1} | R1], [{Id, C2} | R2]) ->
 cmp_lists(_, _) ->
     error(different_chunks).
     
-strip_rel(Root) ->
+strip_rel(Root, AdditionalChunks) ->
     ok = assert_directory(Root),
-    strip_fils(filelib:wildcard(filename:join(Root, "lib/*/ebin/*.beam"))).
+    strip_fils(filelib:wildcard(filename:join(Root, "lib/*/ebin/*.beam")), AdditionalChunks).
 
 %% -> {ok, [{Mod, BinaryOrFileName}]} | throw(Error)
-strip_fils(Files) ->
-    {ok, [begin {ok, Reply} = strip_file(F), Reply end || F <- Files]}.
+strip_fils(Files, AdditionalChunks) ->
+    {ok, [begin {ok, Reply} = strip_file(F, AdditionalChunks), Reply end || F <- Files]}.
 
 %% -> {ok, {Mod, FileName}} | {ok, {Mod, binary()}} | throw(Error)
-strip_file(File) ->
-    {ok, {Mod, Chunks}} = read_significant_chunks(File, significant_chunks()),
+strip_file(File, AdditionalChunks) ->
+    {ok, {Mod, Chunks}} = read_significant_chunks(File, AdditionalChunks ++ significant_chunks()),
     {ok, Stripped0} = build_module(Chunks),
     Stripped = compress(Stripped0),
     case File of
@@ -650,10 +1030,13 @@ chunks_to_data([{abst_chunk, Name} | CNs], Chunks, File, Cs, Module, Atoms, L) -
     {NewAtoms, Ret} =
 	case catch chunk_to_data(debug_info, DbgiChunk, File, Cs, Atoms, Module) of
 	    {DbgiAtoms, {debug_info, {debug_info_v1, Backend, Metadata}}} ->
-		case Backend:debug_info(erlang_v1, Module, Metadata, []) of
+		try Backend:debug_info(erlang_v1, Module, Metadata, []) of
 		    {ok, Code} -> {DbgiAtoms, {abstract_code, {raw_abstract_v1, Code}}};
 		    {error, _} -> {DbgiAtoms, {abstract_code, no_abstract_code}}
-		end;
+                catch
+                    error:undef ->
+                        error({missing_backend,File,Backend})
+                end;
             {error,beam_lib,{key_missing_or_invalid,Path,debug_info}} ->
                 error({key_missing_or_invalid,Path,abstract_code});
 	    _ ->
@@ -668,6 +1051,9 @@ chunks_to_data([{Id, Name} | CNs], Chunks, File, Cs, Module, Atoms, L) ->
 chunks_to_data([], _Chunks, _File, _Cs, Module, _Atoms, L) ->
     {ok, {Module, reverse(L)}}.
 
+chunk_to_data(Id, missing_chunk, _File, _Cs, AtomTable, _Mod) ->
+    %% Missing chunk, only happens when 'allow_missing_chunks' is on.
+    {AtomTable, {Id, missing_chunk}};
 chunk_to_data(attributes=Id, Chunk, File, _Cs, AtomTable, _Mod) ->
     try
 	Term = binary_to_term(Chunk),
@@ -699,6 +1085,18 @@ chunk_to_data(debug_info=Id, Chunk, File, _Cs, AtomTable, Mod) ->
                     {AtomTable, {Id, anno_from_term(Term)}}
 	    end
     end;
+chunk_to_data(documentation=Id, Chunk, File, _Cs, AtomTable, _Mod) ->
+    try
+        case binary_to_term(Chunk) of
+            #docs_v1{} = Term ->
+                {AtomTable, {Id, Term}};
+            _ ->
+                error({invalid_chunk, File, chunk_name_to_id(Id, File)})
+        end
+    catch
+	error:badarg ->
+	    error({invalid_chunk, File, chunk_name_to_id(Id, File)})
+    end;
 chunk_to_data(abstract_code=Id, Chunk, File, _Cs, AtomTable, Mod) ->
     %% Before Erlang/OTP 20.0.
     case Chunk of
@@ -748,6 +1146,7 @@ chunk_name_to_id(attributes, _)      -> "Attr";
 chunk_name_to_id(abstract_code, _)   -> "Abst";
 chunk_name_to_id(debug_info, _)      -> "Dbgi";
 chunk_name_to_id(compile_info, _)    -> "CInf";
+chunk_name_to_id(documentation, _)   -> "Docs";
 chunk_name_to_id(Other, File) -> 
     error({unknown_chunk, File, Other}).
 
@@ -788,8 +1187,7 @@ symbol(_, AT, I1, I2, _I3, _Cnt) ->
     {atm(AT, I1), I2}.
 
 atm(AT, N) ->
-    [{_N, S}] = ets:lookup(AT, N),
-    S.
+    ets:lookup_element(AT, N, 2).
 
 %% AT is updated.
 ensure_atoms({empty, AT}, Cs) ->
@@ -824,10 +1222,8 @@ extract_atom(<<Len, B/binary>>, Encoding) ->
 	     bin :: binary(),
 	     source :: binary() | string()}).
 
-open_file(<<"FOR1",_/binary>>=Binary) ->
-    #bb{bin = Binary, source = Binary};
 open_file(Binary0) when is_binary(Binary0) ->
-    Binary = uncompress(Binary0),
+    Binary = maybe_uncompress(Binary0),
     #bb{bin = Binary, source = Binary};
 open_file(FileName) ->
     case file:open(FileName, [read, raw, binary]) of
@@ -843,7 +1239,7 @@ read_all(Fd, FileName, Bins) ->
 	    read_all(Fd, FileName, [Bin | Bins]);
 	eof ->
 	    ok = file:close(Fd),
-	    #bb{bin = uncompress(reverse(Bins)), source = FileName};
+	    #bb{bin = maybe_uncompress(reverse(Bins)), source = FileName};
 	Error ->
 	    ok = file:close(Fd),
 	    file_error(FileName, Error)
@@ -873,20 +1269,22 @@ beam_filename(Bin) when is_binary(Bin) ->
 beam_filename(File) ->
     filename:rootname(File, ".beam") ++ ".beam".
 
+%% Do not attempt to uncompress if we have the proper .beam format.
+%% This clause matches binaries given as input.
+maybe_uncompress(<<"FOR1",_/binary>>=Binary) ->
+    Binary;
+%% This clause matches the iolist read from files.
+maybe_uncompress([<<"FOR1",_/binary>>|_]=IOData) ->
+    iolist_to_binary(IOData);
+maybe_uncompress(IOData) ->
+    try
+	zlib:gunzip(IOData)
+    catch
+	_:_ -> iolist_to_binary(IOData)
+    end.
 
-uncompress(Binary0) ->
-    {ok, Fd} = ram_file:open(Binary0, [write, binary]),
-    {ok, _} = ram_file:uncompress(Fd),
-    {ok, Binary} = ram_file:get_file(Fd),
-    ok = ram_file:close(Fd),
-    Binary.
-
-compress(Binary0) ->
-    {ok, Fd} = ram_file:open(Binary0, [write, binary]),
-    {ok, _} = ram_file:compress(Fd),
-    {ok, Binary} = ram_file:get_file(Fd),
-    ok = ram_file:close(Fd),
-    Binary.
+compress(IOData) ->
+    zlib:gzip(IOData).
 
 %% -> ok | throw(Error)
 assert_directory(FileName) ->
@@ -909,14 +1307,15 @@ error(Reason) ->
 
 %% The following chunks must be kept when stripping a BEAM file.
 
+-doc false.
 significant_chunks() ->
-    ["Line" | md5_chunks()].
+    ["Line", "Type" | md5_chunks()].
 
 %% The following chunks are significant when calculating the MD5
 %% for a module. They are listed in the order that they should be MD5:ed.
 
 md5_chunks() ->
-    ["Atom", "AtU8", "Code", "StrT", "ImpT", "ExpT", "FunT", "LitT"].
+    ["Atom", "AtU8", "Code", "StrT", "ImpT", "ExpT", "FunT", "LitT", "Meta"].
 
 %% The following chunks are mandatory in every Beam file.
 
@@ -941,7 +1340,7 @@ decrypt_chunk(Type, Module, File, Id, Bin) ->
 	KeyString = get_crypto_key({debug_info, Type, Module, File}),
 	{Type,Key,IVec,_BlockSize} = make_crypto_key(Type, KeyString),
 	ok = start_crypto(),
-	NewBin = crypto:block_decrypt(Type, Key, IVec, Bin),
+	NewBin = crypto:crypto_one_time(des_ede3_cbc, Key, IVec, Bin, false),
 	binary_to_term(NewBin)
     catch
 	_:_ ->
@@ -984,6 +1383,7 @@ start_crypto() ->
 	    ok
     end.
 
+-doc false.
 get_crypto_key(What) ->
     call_crypto_server({get_crypto_key, What}).
 
@@ -1007,6 +1407,7 @@ call_crypto_server_1(Req) ->
     erlang:yield(),
     call_crypto_server(Req).
 
+-doc false.
 -spec init([]) -> {'ok', #state{}}.
 
 init([]) ->
@@ -1016,6 +1417,7 @@ init([]) ->
                | {'crypto_key_fun', _}
                | {'get_crypto_key', _}.
 
+-doc false.
 -spec handle_call(calls(), {pid(), term()}, #state{}) ->
         {'noreply', #state{}} |
 	{'reply', 'error' | {'error','badfun' | 'exists'}, #state{}} |
@@ -1086,32 +1488,37 @@ handle_call(clear_crypto_key_fun, _From, S) ->
 	    {stop,normal,{ok,Result},S}
     end.
 
+-doc false.
 -spec handle_cast(term(), #state{}) -> {'noreply', #state{}}.
 
 handle_cast(_, State) ->
     {noreply, State}.
 
+-doc false.
 -spec handle_info(term(), #state{}) -> {'noreply', #state{}}.
 
 handle_info(_, State) ->
     {noreply, State}.
 
+-doc false.
 -spec code_change(term(), #state{}, term()) -> {'ok', #state{}}.
 
 code_change(_OldVsn, State, _Extra) ->
     {ok, State}.
 
+-doc false.
 -spec terminate(term(), #state{}) -> 'ok'.
 
 terminate(_Reason, _State) ->
     ok.
 
 crypto_key_fun_from_file() ->
+    UserConfig = filename:basedir(user_config,"erlang"),
     case init:get_argument(home) of
 	{ok,[[Home]]} ->
-	    crypto_key_fun_from_file_1([".",Home]);
+	    crypto_key_fun_from_file_1([".", Home, UserConfig]);
 	_ ->
-	    crypto_key_fun_from_file_1(["."])
+	    crypto_key_fun_from_file_1([".", UserConfig])
     end.
 
 crypto_key_fun_from_file_1(Path) ->
